@@ -174,16 +174,40 @@ module NATS
         # Hostname of current server; used for when TLS host
         # verification is enabled.
         @hostname = nil
+        @single_url_connect_used = false
 
         # New style request/response implementation.
         @resp_sub = nil
         @resp_map = nil
         @resp_sub_prefix = nil
         @nuid = NATS::NUID.new
+
+        # NKEYS
+        @user_credentials = nil
+        @nkeys_seed = nil
+        @user_nkey_cb = nil
+        @user_jwt_cb = nil
+        @signature_cb = nil
       end
 
       # Establishes connection to NATS.
-      def connect(opts={})
+      def connect(uri=nil, opts={})
+        case uri
+        when String
+          # Initialize TLS defaults in case any url is using it.
+          srvs = opts[:servers] = process_uri(uri)
+          if srvs.any? {|u| u.scheme == 'tls'} and !opts[:tls]
+            tls_context = OpenSSL::SSL::SSLContext.new
+            tls_context.set_params
+            opts[:tls] = {
+              context: tls_context
+            }
+          end
+          @single_url_connect_used = true if srvs.size == 1
+        when Hash
+          opts = uri
+        end
+
         opts[:verbose] = false if opts[:verbose].nil?
         opts[:pedantic] = false if opts[:pedantic].nil?
         opts[:reconnect] = true if opts[:reconnect].nil?
@@ -226,6 +250,11 @@ module NATS
           class << self; alias_method :request, :old_request; end
         end
 
+        # NKEYS
+        @user_credentials ||= opts[:user_credentials]
+        @nkeys_seed ||= opts[:nkeys_seed]
+        setup_nkeys_connect if @user_credentials or @nkeys_seed
+
         # Check for TLS usage
         @tls = @options[:tls]
 
@@ -245,7 +274,12 @@ module NATS
           @status = CONNECTING
 
           # Use the hostname from the server for TLS hostname verification.
-          @hostname = srv[:hostname]
+          if client_using_secure_connection? and single_url_connect_used?
+            # Always reuse the original hostname used to connect.
+            @hostname ||= srv[:hostname]
+          else
+            @hostname = srv[:hostname]
+          end
 
           # Established TCP connection successfully so can start connect
           process_connect_init
@@ -634,7 +668,8 @@ module NATS
           if connect_urls
             srvs = []
             connect_urls.each do |url|
-              u = URI.parse("nats://#{url}")
+              scheme = client_using_secure_connection? ? "tls" : "nats"
+              u = URI.parse("#{scheme}://#{url}")
 
               # Skip in case it is the current server which we already know
               next if @uri.host == u.host && @uri.port == u.port
@@ -753,6 +788,10 @@ module NATS
         @uri.scheme == "tls" || @tls
       end
 
+      def single_url_connect_used?
+        @single_url_connect_used
+      end
+
       def send_command(command)
         @pending_size += command.bytesize
         @pending_queue << command
@@ -774,13 +813,22 @@ module NATS
         }
         cs[:name] = @options[:name] if @options[:name]
 
-        if auth_connection?
+        case
+        when auth_connection?
           if @uri.password
             cs[:user] = @uri.user
             cs[:pass] = @uri.password
           else
             cs[:auth_token] = @uri.user
           end
+        when @user_credentials
+          nonce = @server_info[:nonce]
+          cs[:jwt] = @user_jwt_cb.call
+          cs[:sig] = @signature_cb.call(nonce)
+        when @nkeys_seed
+          nonce = @server_info[:nonce]
+          cs[:nkey] = @user_nkey_cb.call
+          cs[:sig] = @signature_cb.call(nonce)
         end
 
         "CONNECT #{cs.to_json}#{CR_LF}"
@@ -933,8 +981,13 @@ module NATS
         if !line or line.empty?
           raise ConnectError.new("nats: protocol exception, INFO not received")
         end
-        _, info_json = line.split(' ')
-        process_info(info_json)
+
+        if match = line.match(NATS::Protocol::INFO)
+          info_json = match.captures.first
+          process_info(info_json)
+        else
+          raise ConnectError.new("nats: protocol exception, INFO not valid")
+        end
 
         case
         when (server_using_secure_connection? and client_using_secure_connection?)
@@ -1021,7 +1074,12 @@ module NATS
           @stats[:reconnects] += 1
 
           # Set hostname to use for TLS hostname verification
-          @hostname = srv[:hostname]
+          if client_using_secure_connection? and single_url_connect_used?
+            # Reuse original hostname name in case of using TLS.
+            @hostname ||= srv[:hostname]
+          else
+            @hostname = srv[:hostname]
+          end
 
           # Established TCP connection successfully so can start connect
           process_connect_init
@@ -1241,6 +1299,115 @@ module NATS
           connect_timeout: DEFAULT_CONNECT_TIMEOUT
         })
       end
+
+      def setup_nkeys_connect
+        begin
+          require 'nkeys'
+          require 'base64'
+        rescue LoadError
+          raise(Error, "nkeys is not installed")
+        end
+
+        case
+        when @nkeys_seed
+          @user_nkey_cb = proc {
+            seed = File.read(@nkeys_seed).chomp
+            kp = NKEYS::from_seed(seed)
+
+            # Take a copy since original will be gone with the wipe.
+            pub_key = kp.public_key.dup
+            kp.wipe!
+
+            pub_key
+          }
+
+          @signature_cb = proc { |nonce|
+            seed = File.read(@nkeys_seed).chomp
+            kp = NKEYS::from_seed(seed)
+            raw_signed = kp.sign(nonce)
+            kp.wipe!
+            encoded = Base64.urlsafe_encode64(raw_signed)
+            encoded.gsub('=', '')
+          }
+        when @user_credentials
+          # When the credentials are within a single decorated file.
+          @user_jwt_cb = proc {
+            jwt_start = "BEGIN NATS USER JWT".freeze
+            found = false
+            jwt = nil
+            File.readlines(@user_credentials).each do |line|
+              case
+              when found
+                jwt = line.chomp
+                break
+              when line.include?(jwt_start)
+                found = true
+              end
+            end
+            raise(Error, "No JWT found in #{@user_credentials}") if not found
+
+            jwt
+          }
+
+          @signature_cb = proc { |nonce|
+            seed_start = "BEGIN USER NKEY SEED".freeze
+            found = false
+            seed = nil
+            File.readlines(@user_credentials).each do |line|
+              case
+              when found
+                seed = line.chomp
+                break
+              when line.include?(seed_start)
+                found = true
+              end
+            end
+            raise(Error, "No nkey user seed found in #{@user_credentials}") if not found
+
+            kp = NKEYS::from_seed(seed)
+            raw_signed = kp.sign(nonce)
+
+            # seed is a reference so also cleared when doing wipe,
+            # which can be done since Ruby strings are mutable.
+            kp.wipe
+            encoded = Base64.urlsafe_encode64(raw_signed)
+
+            # Remove padding
+            encoded.gsub('=', '')
+          }
+        end
+      end
+
+      def process_uri(uris)
+        connect_uris = []
+        uris.split(',').each do |uri|
+          opts = {}
+
+          # Scheme
+          if uri.include?("://")
+            scheme, uri = uri.split("://")
+            opts[:scheme] = scheme
+          else
+            opts[:scheme] = 'nats'
+          end
+
+          # UserInfo
+          if uri.include?("@")
+            userinfo, endpoint = uri.split("@")
+            host, port = endpoint.split(":")
+            opts[:userinfo] = userinfo
+          else
+            host, port = uri.split(":")
+          end
+
+          # Host and Port
+          opts[:host] = host || "localhost"
+          opts[:port] = port || DEFAULT_PORT
+
+          connect_uris << URI::Generic.build(opts)
+        end
+        connect_uris
+      end
     end
 
     # Implementation adapted from https://github.com/redis/redis-rb
